from bisect import bisect_left
import math
from typing import Tuple, Union, List, Dict, Callable
from numbers import Number
import random
import numpy as np
import matplotlib.pyplot as plt

from ..parameterization._parameter_space import ParameterSpace
from ..perturbations._param_values import ParamPerturbation
from ._discretization import Discretization
from ..exceptions import DimensionalityException
from ..parameters import Parameter
from .._state import State, ParameterSpaceState
from .._utils_1d import (
    interpolate_depth_profile, 
    compute_voronoi1d_cell_extents, 
    insert_scalar,
    nearest_index,
    delete
)


SQRT_TWO_PI = math.sqrt(2 * math.pi)


class Voronoi(Discretization):
    r"""Utility class for Voronoi tessellation

    Parameters
    ----------
    name : str
        name attributed to the Voronoi tessellation, for display and storing 
        purposes
    spatial_dimensions : int
        number of dimensions of the desired Voronoi tessellation, e.g. 1D,
        2D, or 3D.
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the Voronoi
        sites in each dimension. 
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min``-``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range `n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if 
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and 
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is
            
            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)
            
    parameters : List[Parameter], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the free parameters associated with the newborn 
        Voronoi cell by randomly drawing from their prior or by perturbing the 
        value found in the nearest Voronoi cell (default).
    """
    def __init__(
        self,
        name: str,
        spatial_dimensions: Number,
        vmin: Union[Number, np.ndarray],
        vmax: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        n_dimensions: int = None, 
        n_dimensions_min: int = 1, 
        n_dimensions_max: int = 10, 
        n_dimensions_init_range: Number = 0.3, 
        parameters: List[Parameter] = None, 
        birth_from: str = "neighbour",  # either "neighbour" or "prior"
    ):
        super().__init__(
            name=name,
            spatial_dimensions=spatial_dimensions,
            perturb_std=perturb_std,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters,
            birth_from=birth_from,
            vmin=vmin,
            vmax=vmax
        )
        self.vmin = vmin
        self.vmax = vmax
        msg = "The %s number of Voronoi cells, "
        if n_dimensions is not None:
            assert n_dimensions > 0, msg % "minimum" + "`n_dimensions`, should be greater than zero"
            assert isinstance(n_dimensions, int), msg % "minimum" + "`n_dimensions`, should be an integer"
            assert isinstance(n_dimensions, int), msg % "maximum" + "`n_dimensions`, should be an integer"
            
    def log_prior(self, *args):
        r"""
        BayesBay implements the grid trick, which calculates the prior 
        probability of a Voronoi discretization through the combinatorial 
        formula :math:`{N \choose k}^{-1}`, with `k` denoting the number of 
        Voronoi sites and `N` the number of possible positions allowed for the 
        sites [3]_.
        
        References
        ----------
        .. [3] Bodin and Sambridge (2009), Seismic tomography with the reversible 
            jump algorithm
        """
        raise NotImplementedError

    def _init_perturbation_funcs(self):
        ParameterSpace._init_perturbation_funcs(self)
        self._perturbation_funcs.append(ParamPerturbation(self.name, [self]))

    @property
    def perturbation_functions(self) -> List[Callable[[State], Tuple[State, Number]]]:
        r"""the list of perturbation functions allowed in the parameter space linked to
        the Voronoi discretization. Each function takes in a state (see :class:`State`) 
        and returns a new state along with the corresponding partial acceptance 
        probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        """
        return self._perturbation_funcs


class Voronoi1D(Voronoi):
    r"""Utility class for Voronoi tessellation in 1D

    Parameters
    ----------
    name : str
        name attributed to the Voronoi tessellation, for display and storing 
        purposes
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the Voronoi
        sites in each dimension. 
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min``-``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range ``n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if 
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and 
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is::
            
            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)
            
    parameters : List[Parameter], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the free parameters associated with the newborn 
        Voronoi cell by randomly drawing from their prior or by perturbing the 
        value found in the nearest Voronoi cell (default).
    """
    def __init__(        
            self,
            name: str,
            vmin: Number,
            vmax: Number,
            perturb_std: Union[Number, np.ndarray],
            n_dimensions: int = None, 
            n_dimensions_min: int = 1, 
            n_dimensions_max: int = 10, 
            n_dimensions_init_range: Number = 0.3, 
            parameters: List[Parameter] = None, 
            birth_from: str = "neighbour"  # either "neighbour" or "prior"
        ): 
        super().__init__(
            name=name,
            spatial_dimensions=1,
            vmin=vmin,
            vmax=vmax,
            perturb_std=perturb_std,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters,
            birth_from=birth_from
        )
    
    def initialize(self) -> ParameterSpaceState:
        """initializes the parameter space linked to the Voronoi tessellation

        Returns
        -------
        ParameterSpaceState
            an initial parameter space state
        """
        # initialize number of dimensions
        if not self.trans_d:
            n_voronoi_cells = self._n_dimensions
        else:
            init_range = self._n_dimensions_init_range
            n_dims_min = self._n_dimensions_min
            n_dims_max = self._n_dimensions_max
            init_max = int((n_dims_max - n_dims_min) * init_range + n_dims_min)
            n_voronoi_cells = random.randint(n_dims_min, init_max)
        lb, ub = self.vmin, self.vmax
        voronoi_sites = np.sort(np.random.uniform(lb, ub, n_voronoi_cells))

        # initialize parameter values
        parameter_vals = {"discretization": voronoi_sites}
        for name, param in self.parameters.items():
            parameter_vals[name] = param.initialize(voronoi_sites)
        return ParameterSpaceState(n_voronoi_cells, parameter_vals)    
    
    def _perturb_site(self, site: Number) -> Number:
        """perturbes a Voronoi  site
        
        Parameters
        ----------
        site : float
            Voronoi site position

        Returns
        -------
        Number
            perturbed Voronoi site position
        """        
        while True:
            random_deviate = random.normalvariate(0, self.perturb_std)
            new_site = site + random_deviate
            if new_site >= self.vmin and new_site <= self.vmax:
                return new_site   
        
    def perturb_value(self, old_ps_state: ParameterSpaceState, isite: Number):
        r"""perturbs the value of one Voronoi site and calculates the log of the
        partial acceptance probability
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.

        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            the current parameter space state
        isite : Number
            the index of the Voronoi site to be perturbed

        Returns
        -------
        Tuple[ParameterSpaceState, Number]
            the new parameter space state and its associated partial acceptance 
            probability excluding log likelihood ratio
        """
        old_sites = old_ps_state["discretization"]
        old_site = old_sites[isite]
        new_site = self._perturb_site(old_sites[isite])
        new_sites = old_sites.copy()
        new_sites[isite] = new_site
        isort = np.argsort(new_sites)
        new_sites = new_sites[isort]
        new_values = {"discretization": new_sites}
        log_prior_ratio = 0
        for param_name, param in self.parameters.items():
            values = old_ps_state[param_name]
            if param.position is not None:
                log_prior_old = param.log_prior(values[isite], old_site)
                log_prior_new = param.log_prior(values[isite], new_site)
                log_prior_ratio += log_prior_new - log_prior_old
            new_values[param_name] = values[isort]
            
        new_ps_state = ParameterSpaceState(old_ps_state.n_dimensions, new_values)
        return new_ps_state, log_prior_ratio # log_proposal_ratio=0 and log_det_jacobian=0
        
    def _initialize_newborn_params(
            self, 
            new_site: Number, 
            old_sites: np.ndarray, 
            param_space_state: ParameterSpaceState
            ):
        """initialize the parameter values in the newborn dimension
    
        Parameters
        ----------
        new_site : Number
            position of the newborn Voronoi site
        old_sites : np.ndarray
            all positions of the current Voronoi sites
        param_space_state : ParameterSpaceState
            current parameter space state
    
        Returns
        -------
        Dict[str, float]
            key value pairs that map parameter names to values of the ``new_site``
        """
        if self.birth_from == 'prior':
            return self._initialize_params_from_prior(
                new_site, old_sites, param_space_state
                )
        return self._initialize_params_from_neighbour(
            new_site, old_sites, param_space_state
            )
    
    def _initialize_params_from_prior(
            self, 
            new_site: Number, 
            old_sites: np.ndarray, 
            param_space_state: ParameterSpaceState
            ) -> Tuple[Dict[str, np.ndarray], None]:
        """initialize the newborn dimension by randomly drawing parameter values
        from the prior
    
        Parameters
        ----------
        new_site : Number
            position of the newborn Voronoi site
        old_sites : np.ndarray
            all positions of the current Voronoi sites
        param_space_state : ParameterSpaceState
            current parameter space state
    
        Returns
        -------
        Tuple[Dict[str, np.ndarray], None]
            key value pairs that map parameter names to values of the ``new_site``
        """
        new_born_values = dict()
        for param_name, param in self.parameters.items():
            new_value = param.sample(new_site)
            new_born_values[param_name] = new_value
        return new_born_values, None
    
    def _initialize_params_from_neighbour(
        self, 
        new_site: Number, 
        old_sites: np.ndarray, 
        param_space_state: ParameterSpaceState
        ) -> Tuple[Dict[str, np.ndarray], Number]:
        """initialize the newborn parameter values by perturbing the nearest 
        Voronoi cell
    
        Parameters
        ----------
        new_site : Number
            position of the newborn Voronoi cell
        old_sites : np.ndarray
            all positions of the current Voronoi cells
        param_space_state : ParameterSpaceState
            current parameter space state
    
        Returns
        -------
        Tuple[Dict[str, np.ndarray], Number]
            key value pairs that map parameter names to values of the ``new_site``
            and the index of the Voronoi neighbour
        """
        isite = nearest_index(xp=new_site, x=old_sites, xlen=old_sites.size)
        new_born_values = dict()
        for param_name, param in self.parameters.items():
            old_values = param_space_state[param_name]
            new_value, _ = param.perturb_value(old_values[isite], new_site)
            new_born_values[param_name] = new_value
        return new_born_values, isite
    
    def _log_probability_ratio_birth(
            self, 
            old_isite: Number, 
            old_ps_state: ParameterSpaceState, 
            new_isite: Number, 
            new_ps_state: ParameterSpaceState
            ):
        if self.birth_from == 'prior':
            return 0
        return self._log_probability_ratio_birth_from_neighbour(
            old_isite, old_ps_state, new_isite, new_ps_state
        )
    
    def _log_probability_ratio_birth_from_neighbour(
            self, 
            old_isite: Number, 
            old_ps_state: ParameterSpaceState, 
            new_isite: Number, 
            new_ps_state: ParameterSpaceState
    ):
        old_site = old_ps_state["discretization"][old_isite]
        new_site = new_ps_state["discretization"][new_isite]
        log_prior_ratio = 0
        log_proposal_ratio = 0
        for param_name, param in self.parameters.items():
            new_value = new_ps_state[param_name][new_isite]
            log_prior_ratio += param.log_prior(new_value, new_site)
            
            old_value = old_ps_state[param_name][old_isite]
            perturb_std = param.get_perturb_std(old_site)
            log_proposal_ratio += (
                math.log(perturb_std * SQRT_TWO_PI)
                + (new_value - old_value) ** 2 / (2 * perturb_std**2)
            )
        return log_prior_ratio + log_proposal_ratio # log_det_jacobian is 1          
    
    def birth(self, old_ps_state: ParameterSpaceState) -> Tuple[ParameterSpaceState, float]:
        r"""creates a new Voronoi cell, initializes all free parameters 
        associated with it, and returns the pertubed state along with the
        log of the corresponding partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
    
        In this case, the prior probability of the model :math:`{\bf m}` is
        
        .. math::
            p({\bf m}) = p({\bf c} \mid k) p(k) \prod_i{p({\bf v}_i \mid {\bf c})} ,
        
        where :math:`k` denotes the number of Voronoi cells, each entry of the 
        vector :math:`{\bf c}` corresponds to the position of a Voronoi site, 
        and each :math:`i`\ th free parameter :math:`{\bf v}` has the same 
        dimensionality as :math:`{\bf c}`. 
        
        Following [1]_, :math:`p({\bf c} \mid k) = \frac{k! \left(N - k \right)!}{N!}`. If we then
        assume that :math:`p(k) = \frac{1}{\Delta k}`, where :math:`\Delta k = k_{max} - k_{min}`,
        the prior ratio reads
        
        .. math::
            \frac{p({\bf m'})}{p({\bf m})} = 
            \frac{(k+1) \prod_i p(v_i^{k+1})}{(N-k)},
                                         
        where :math:`p(v_i^{k+1})` denotes the prior probability of the newly
        born :math:`i`\ th parameter, which may be dependent on :math:`{\bf c}`.
        The proposal ratio reads
        
        .. math::
            \frac{q({\bf m} \mid {\bf m'})}{q({\bf m'} \mid {\bf m})} =
            \frac{(N-k)}{(k+1) \prod_i q_{v_i}^{k+1}},
                         
        where :math:`q_{v_i}^{k+1}` denotes the proposal probability for the
        newly born :math:`i`\ th parameter in the new dimension. It is easy to
        show that, in the case of a birth from neighbor [1]_ or a birth from
        prior [2]_ (see :attr:`birth_from`), :math:`\lvert \mathbf{J} \rvert = 1`
        and :math:`\alpha_{p} = \frac{p({\bf m'})}{p({\bf m})} \frac{q({\bf m} \mid {\bf m'})}{q({\bf m'} \mid {\bf m})}`. 
        It follows that
        
        .. math::
            \alpha_{p} = 
            \frac{(k+1) \prod_i p(v_i^{k+1})}{(N-k)} \frac{(N-k)}{(k+1) \prod_i q_{v_i}^{k+1}} = 
            \frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}}.
            
        In the case of a birth from prior, :math:`q_{v_i}^{k+1} = p(v_i^{k+1})`
        and
        
        .. math::
            \alpha_{p} = 
            \frac{\prod_i p(v_i^{k+1})}{\prod_i{p(v_i^{k+1})}} = 1.
                                                                  
        In the case of a birth from neighbor, :math:`q_{v_i}^{k+1} = 
        \frac{1}{\theta \sqrt{2 \pi}} \exp \lbrace -\frac{\left( v_i^{k+1} - v_i \right)^2}{2\theta^2} \rbrace`,
        where the newly born value, :math:`v_i^{k+1}`, is generated by perturbing
        the original value, :math:`v_i`, of the :math:`i`\ th parameter. This is 
        achieved through a random deviate from the normal distribution 
        :math:`\mathcal{N}(v_i, \theta)`, with :math:`\theta` denoting the 
        standard deviation of the Gaussian used to carry out the perturbation
        (see, for example, :attr:`bayesbay.parameters.UniformParameter.perturb_std`) . 
        The partial acceptance probability is then computed numerically.
                  
    
        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            current parameter space state
    
        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`log(\alpha_{p}) = \log(\frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}})`
            
        References
        ----------
        .. [1] Bodin et al. 2012, Transdimensional inversion of receiver functions 
            and surface wave dispersion
        .. [2] Hawkins and Sambridge 2015, Geophysical imaging using trans-dimensional 
            trees
        """
        # prepare for birth perturbation
        n_cells = old_ps_state.n_dimensions
        if n_cells == self._n_dimensions_max:
            raise DimensionalityException("Birth")
        # randomly choose a new Voronoi site position
        lb, ub = self.vmin, self.vmax
        new_site = random.uniform(lb, ub)
        old_sites = old_ps_state["discretization"]
        unsorted_values, i_nearest = self._initialize_newborn_params(
            new_site, old_sites, old_ps_state
        )
        new_values = dict()
        idx_insert = bisect_left(old_sites, new_site)
        new_sites = insert_scalar(old_sites, idx_insert, new_site)
        new_values["discretization"] = new_sites
        for name, value in unsorted_values.items():
            old_values = old_ps_state[name]
            new_values[name] = insert_scalar(old_values, idx_insert, value)
        new_ps_state = ParameterSpaceState(n_cells + 1, new_values)
        return new_ps_state, self._log_probability_ratio_birth(
            i_nearest, old_ps_state, idx_insert, new_ps_state
        )
    
    def _log_probability_ratio_death(
            self, 
            iremove: Number, 
            old_ps_state: ParameterSpaceState, 
            new_ps_state: ParameterSpaceState
            ):
        if self.birth_from == 'prior':
            return 0
        return self._log_probability_ratio_death_from_neighbour(
            iremove, old_ps_state, new_ps_state
        )
    
    def _log_probability_ratio_death_from_neighbour(
            self, 
            iremove: Number, 
            old_ps_state: ParameterSpaceState, 
            new_ps_state: ParameterSpaceState
            ):
        old_sites = old_ps_state["discretization"]
        new_sites = new_ps_state["discretization"]
        i_nearest = nearest_index(
            xp=old_sites[iremove], x=new_sites, xlen=new_sites.size
        )
        return -self._log_probability_ratio_birth(
            i_nearest, new_ps_state, iremove, old_ps_state
        )

    def death(self, old_ps_state: ParameterSpaceState):
        r"""removes a new Voronoi cell and returns the pertubed state along with 
        the log of the corresponding partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
    
        It is straightforward to show that this equals the reciprocal of
        the partial acceptance probability obtained in the case of a birth
        perturbation (see :meth:`birth`), i.e.,
                
        .. math::
            \alpha_{p} = \frac{\prod_i{q_{v_i}^{k+1}}}{\prod_i p(v_i^{k+1})}.
    
        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            current parameter space state
    
        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`log(\alpha_{p}) = -\log(\frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}})`
        """
        # prepare for death perturbation
        n_cells = old_ps_state.n_dimensions
        if n_cells == self._n_dimensions_min:
            raise DimensionalityException("Death")
        # randomly choose an existing Voronoi site to kill
        iremove = random.randint(0, n_cells - 1)
        # remove parameter values for the removed site
        new_values = dict()
        for name, old_values in old_ps_state.param_values.items():
            new_values[name] = delete(old_values, iremove)
        new_ps_state = ParameterSpaceState(n_cells - 1, new_values) 
        return new_ps_state, self._log_probability_ratio_death(
            iremove, old_ps_state, new_ps_state
        )
    
    @staticmethod
    def compute_cell_extents(voronoi_sites: np.ndarray, lb=0, ub=-1, fill_value=0):
        r"""compute Voronoi cell extents from the Voronoi sites. Voronoi-cell
        boundaries are first drawn at the midpoint between consecutive Voronoi
        nuclei. The extent is then derived from the distance between consecutive
        boundaries.

        Parameters
        ----------
        voronoi_sites : np.ndarray of shape (n,)
            Voronoi-site positions. These should be greater or equal to zero

        lb, ub : float
            Lower and upper bounds used in the calculation of Voronoi-cell
            extents. Negative values for `lb` or `ub` denote an unbounded cell.
            The extent of an unbounded cell is set to `fill_value`

        fill_value : float
            Value attributed to unbounded Voronoi cells

        Returns
        -------
        np.ndarray
            Voronoi-cell extents

        Examples
        --------
        >>> depth = np.array([2, 5.5, 8, 10])

        >>> Voronoi1D.compute_cell_extents(depth, lb=0, ub=-1, fill_value=np.nan)
        array([3.75, 3.  , 2.25,  nan])

        >>> Voronoi1D.compute_cell_extents(depth, lb=-1, ub=-1, fill_value=np.nan)
        array([ nan, 3.  , 2.25,  nan])

        >>> Voronoi1D.compute_cell_extents(depth, lb=0, ub=15, fill_value=np.nan)
        array([3.75, 3.  , 2.25, 6.  ])
        """
        return compute_voronoi1d_cell_extents(
            voronoi_sites, lb=lb, ub=ub, fill_value=fill_value
        )

    @staticmethod
    def plot_depth_profiles_density(
        samples_voronoi_cell_extents: np.ndarray,
        samples_param_values: np.ndarray,
        depths_bins: Union[int, np.ndarray] = 100, 
        param_values_bins: Union[int, np.ndarray] = 100,
        ax=None,
        **kwargs,
    ):
        """plot a 2D histogram of parameter values density at refined depth positions

        Parameters
        ----------
        samples_voronoi_cell_extents : ndarray
            A 2D numpy array where each row represents a sample of thicknesses (or
            Voronoi cell extents)
        samples_param_values : ndarray
            A 2D numpy array where each row represents a sample of parameter values
            (e.g., velocities)
        depths_bins: int or np.ndarray, optional
            The depth bins or their number, default to 100
        param_values_bins: int or np.ndarray, optional
            The parameter value bins or their number, default to 100
        ax : Axes, optional
            An optional Axes object to plot on
        kwargs : dict, optional
            Additional keyword arguments to pass to ax.hist2d

        Returns
        -------
        ax : Axes
            The Axes object containing the 2D histogram
        cbar: Colorbar
            The Colorbar object associated with the 2D histogram
        
        Examples
        --------
        .. code-block:: python
        
            from bayesbay.discretization import Voronoi1D
            
            # define and run the Bayesian inversion
            ...
            
            # plot
            results = inversion.get_results()
            samples_voronoi_cell_extents = [
                Voronoi1D.compute_cell_extents(d) for d in results["my_voronoi.discretization"]
            ]
            samples_param_values = results["vs"]
            ax = Voronoi1D.plot_depth_profiles_density(
                samples_voronoi_cell_extents, samples_param_values
            )
        """
        if ax is None:
            _, ax = plt.subplots()
        
        if isinstance(depths_bins, int):
            depths = []
            for cell_extents in samples_voronoi_cell_extents:
                depths.extend(np.cumsum(np.array(cell_extents)))
            depth_max = np.max(depths)
            depth_min = 0
            new_depths = np.linspace(depth_min, depth_max, depths_bins)
        elif isinstance(depths_bins, np.ndarray):
            new_depths = depths_bins
        else:
            raise TypeError("depths_bins should be either an int or np.ndarray")

        new_param_values = Voronoi1D._interpolate_depth_profiles(
            samples_voronoi_cell_extents, samples_param_values, new_depths
        )
        
        # plotting the 2D histogram
        cax = ax.hist2d(
            new_param_values.ravel(), 
            np.tile(new_depths, new_param_values.shape[0]), 
            bins=(param_values_bins, len(new_depths)), 
            **kwargs
        )
        # colorbar (for the histogram density)
        cbar = plt.colorbar(cax[3], ax=ax, aspect=35, pad=0.02)
        cbar.set_label("Counts")
        if ax.get_ylim()[0] < ax.get_ylim()[1]:
            ax.invert_yaxis()
        ax.set_xlabel("Parameter values")
        ax.set_ylabel("Depth")
        return ax, cbar

    @staticmethod
    def plot_interface_hist(
        samples_voronoi_cell_extents,
        bins=100,
        ax=None,
        **kwargs,
    ):
        """plot the 1D depth histogram of interfaces

        Parameters
        ----------
        samples_voronoi_cell_extents : list
            a list of voronoi cell extents (thicknesses in the 1D case)
        bins : int, optional
            number of vertical bins, by default 100
        ax : matplotlib.axes.Axes, optional
            an optional user-provided ax, by default None

        Returns
        -------
        matplotlib.axes.Axes
            the resulting plot that has the depth distribution on it
        """
        if ax is None:
            _, ax = plt.subplots()
        depths = []
        for thicknesses in samples_voronoi_cell_extents:
            depths.extend(np.cumsum(thicknesses))
        # calculate 1D histogram
        h, e = np.histogram(depths, bins=bins, density=True)
        # plot the histogram
        ax.barh(e[:-1], h, height=np.diff(e), align="edge", label="histogram", **kwargs)
        if ax.get_ylim()[0] < ax.get_ylim()[1]:
            ax.invert_yaxis()
        ax.set_xlabel("Probability density")
        ax.set_ylabel("Depth")
        return ax

    @staticmethod
    def plot_depth_profile(
        voronoi_cell_extents: list,
        param_values: list,
        ax=None,
        max_depth=None,
        **kwargs,
    ):
        """plot multiple 1D Earth models based on sampled parameters.

        Parameters
        ----------
        voronoi_cell_extents : list
            Voronoi cell extents (thickness of a 1D layered model)
        param_values : ndarray
            parameter values (e.g., velocity) associated with each Voronoi cell
        ax : Axes, optional
            an optional Axes object to plot on
        kwargs : dict, optional
            additional keyword arguments to pass to ax.step

        Returns
        -------
        ax : Axes
            The Axes object containing the plot
        """
        if ax is None:
            _, ax = plt.subplots()

        # Default plotting style for samples
        sample_style = {
            "linewidth": kwargs.pop("linewidth", kwargs.pop("lw", 0.5)),
            "alpha": kwargs.pop("alpha", 1),
            "color": kwargs.pop(
                "color", kwargs.pop("c", "blue")
            ),  # Fixed color for the sample lines
        }
        sample_style.update(kwargs)  # Override with any provided kwargs
        
        depths = np.cumsum(voronoi_cell_extents[:-1])
        if max_depth is not None:
            assert max_depth > depths[-1], ("`max_depth` should be greater"
                                            " than the sum of Voronoi site"
                                            f" cell extents (here, {depths[-1]})")
            final_depth = max_depth
        else:
            final_depth = depths[-1] + max(voronoi_cell_extents)
            
        y = np.insert(np.append(depths, final_depth), 0, 0)
        x = np.insert(param_values, 0, param_values[0])
        ax.step(x, y, where="post", **sample_style)
        if ax.get_ylim()[0] < ax.get_ylim()[1]:
            ax.invert_yaxis()
        if not ax.get_xlabel():
            ax.set_xlabel("Parameter values")
        if not ax.get_ylabel():
            ax.set_ylabel("Depth")
        return ax

    @staticmethod
    def plot_depth_profiles(
        samples_voronoi_cell_extents: list,
        samples_param_values: list,
        ax=None,
        max_depth=None,
        **kwargs,
    ):
        """plot multiple 1D Earth models based on sampled parameters.

        Parameters
        ----------
        samples_voronoi_cell_extents : list
            a list of voronoi cell extents (thicknesses of 1D layered models)
        samples_param_values : ndarray
            a 2D numpy array where each row contains the parameter values 
            (e.g., velocity) associated with the Voronoi discretization found
            in ``samples_voronoi_cell_extents`` at the same row index
        ax : Axes, optional
            an optional Axes object to plot on
        kwargs : dict, optional
            additional keyword arguments to pass to ax.step

        Returns
        -------
        ax : Axes
            The Axes object containing the plot
        """
        if ax is None:
            _, ax = plt.subplots()
        if max_depth is not None:
            ax.set_ylim(0, max_depth)
        # Default plotting style for samples
        sample_style = {
            "linewidth": kwargs.pop("linewidth", kwargs.pop("lw", 0.5)),
            "alpha": kwargs.pop("alpha", 0.2),
            "color": kwargs.pop(
                "color", kwargs.pop("c", "blue")
            ),  # Fixed color for the sample lines
        }
        sample_style.update(kwargs)  # Override with any provided kwargs

        for thicknesses, values in zip(
            samples_voronoi_cell_extents, samples_param_values
        ):
            Voronoi1D.plot_depth_profile(thicknesses, 
                                         values, 
                                         **sample_style, 
                                         ax=ax)

        if ax.get_ylim()[0] < ax.get_ylim()[1]:
            ax.invert_yaxis()
        if not ax.get_xlabel():
            ax.set_xlabel("Parameter values")
        if not ax.get_ylabel():
            ax.set_ylabel("Depth")

        return ax
    
    def interpolate_depth_profile(
        voronoi_cell_extents, param_values, interp_positions
    ):
        """interpolates the values of a physical parameter that is a function of depth
        as defined by the Voronoi discretization onto the specified depth positions

        Parameters
        ----------
        voronoi_cell_extents : np.ndarray
            the extent of each Voronoi cell
        param_values : np.ndarray
            the physical parameter value associated with each Voronoi cell
        interp_positions : np.ndarray
            the depths at which the parameter values will be returned

        Returns
        -------
        np.ndarray
            the physical parameter values associated with ``interp_positions``
        """ 
        return interpolate_depth_profile(
            np.array(voronoi_cell_extents),
            np.array(param_values), 
            interp_positions, 
        )

    @staticmethod
    def _interpolate_depth_profiles(
        samples_voronoi_cell_extents, samples_param_values, interp_positions
    ):
        interp_params = np.zeros((len(samples_param_values), len(interp_positions)))
        for i, (sample_extents, sample_values) in enumerate(
            zip(samples_voronoi_cell_extents, samples_param_values)
        ):
            interp_params[i, :] = interpolate_depth_profile(
                np.array(sample_extents), np.array(sample_values), interp_positions
            )
        return interp_params

    @staticmethod
    def get_depth_profiles_statistics(
        samples_voronoi_cell_extents: list,
        samples_param_values: list,
        interp_positions: np.ndarray,
        percentiles: tuple = (10, 90),
    ) -> dict:
        """get the mean, median, std and percentiles of the given ensemble

        Parameters
        ----------
        samples_voronoi_cell_extents : list
            a list of voronoi cell extents (thicknesses in the 1D case)
        samples_param_values : list
            a list of physical parameter values to draw statistics from
        interp_positions : np.ndarray
            points to interpolate
        percentiles : tuple, optional
            percentiles to calculate, by default (10, 90)

        Returns
        -------
        dict
            a dictionary with these keys: "mean", "median", "std" and "percentile"
        """
        interp_params = Voronoi1D._interpolate_depth_profiles(
            samples_voronoi_cell_extents, samples_param_values, interp_positions
        )
        statistics = {
            "mean": np.mean(interp_params, axis=0),
            "median": np.median(interp_params, axis=0),
            "std": np.std(interp_params, axis=0),
            "percentiles": np.percentile(interp_params, percentiles, axis=0),
        }
        return statistics

    @staticmethod
    def plot_depth_profiles_statistics(
        samples_voronoi_cell_extents: list,
        samples_param_values: list,
        interp_positions: np.ndarray,
        percentiles=(10, 90),
        ax=None,
    ):
        """plot the mean, median, std and percentiles from the given samples

        Parameters
        ----------
        samples_voronoi_cell_extents : list
            a list of voronoi cell extents (thicknesses in the 1D case)
        samples_param_values : list
            a list of physical parameter values to draw statistics from
        interp_positions : _type_
            points to interpolate
        percentiles : tuple, optional
            percentiles to calculate, by default (10, 90)
        ax : matplotlib.axes.Axes, optional
            an optional user-provided ax, by default None

        Returns
        -------
        matplotlib.axes.Axes
            the resulting plot that has the statistics on it
        """
        statistics = Voronoi1D.get_depth_profiles_statistics(
            samples_voronoi_cell_extents,
            samples_param_values,
            interp_positions,
            percentiles,
        )
        mean = statistics["mean"]
        std = statistics["std"]
        percentiles = statistics["percentiles"]
        if ax is None:
            _, ax = plt.subplots()
        ax.plot(mean, interp_positions, color="b", label="Mean")
        ax.plot(mean - std, interp_positions, "b--")
        ax.plot(mean + std, interp_positions, "b--")
        ax.plot(statistics["median"], interp_positions, color="orange", label="Median")
        ax.plot(
            percentiles[0],
            interp_positions,
            color="orange",
            ls="--",
        )
        ax.plot(
            percentiles[1],
            interp_positions,
            color="orange",
            ls="--",
        )
        ax.legend()
        return ax



from ._exceptions import (
    InitException,
    ForwardException,
    DimensionalityException,
    UserFunctionException,
)


__all__ = [
    "InitException",
    "ForwardException",
    "DimensionalityException",
    "UserFunctionException",
]



from typing import List, Callable, Tuple, Any, Dict, Union
from numbers import Number
from collections import defaultdict
from pprint import pformat

from ._markov_chain import MarkovChain, BaseMarkovChain
from .samplers import VanillaSampler, Sampler
from .parameterization import Parameterization
from ._utils import _preprocess_func
from ._log_likelihood import LogLikelihood
from ._target import Target


class BaseBayesianInversion:
    r"""
    A low-level class for Bayesian sampling based on Markov chain Monte Carlo (McMC).

    This class provides the basic structure for setting up and running McMC sampling,
    given user-provided definition of perturbations, likelihood functions and the
    initialization of walkers. At each iteration of the inference process,
    the current model :math:`\bf m` is perturbed to produce :math:`\bf m'`, and
    the new model is accepted with probability

    .. math::

        \alpha({\bf m' \mid m}) = \mbox{min} \Bigg[1,
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{p\left({\bf d}_{obs} \mid {\bf m'}\right)}{p\left({\bf d}_{obs} \mid {\bf m}\right)}}_{\text{Likelihood ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}
        \Bigg],

    where :math:`{\bf d}_{obs}` denotes the observed data and
    :math:`\mathbf{J}` the Jacobian of the transformation.

    Parameters
    ----------
    walkers_starting_states: List[Any]
        a list of starting states for each chain. The states can be of any type so long
        as they are consistent with what is accepted as arguments in the perturbation
        functions and probability functions. The length of this list must be equal to
        the number of chains, i.e. ``n_chains``
    perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]]
        a list of perturbation functions. Each perturbation function should take
        in a model :math:`\mathbf{m}` (any type is allowed, as long as it is
        consistent with ``walkers_starting_states`` and the below functions) and
        perturb it to produce the new model :math:`\bf m'`. Each perturbation function
        should return :math:`\bf m'` along with :math:`\log(
        \frac{p({\bf m'})}{p({\bf m})}
        \frac{q\left({\bf m}
        \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
        \lvert \mathbf{J} \rvert)`, which is used in the calculation of
        the acceptance probability.
    log_like_ratio_func: Union[LogLikelihood, Callable[[Any, Any], Number]], optional
        the log likelihood ratio function :math:`\log(\frac{p(\mathbf{d}_{obs} \mid 
        \mathbf{m'})} {p(\mathbf{d}_{obs} \mid \mathbf{m})})`. It takes the current and 
        proposed models, :math:`\mathbf{m}` and :math:`\mathbf{m'}`, whose type should 
        be consistent with the other arguments of this class, and returns a scalar 
        corresponding to the log likelihood ratio. This is utilised in the calculation 
        of the acceptance probability. If None, ``log_like_func`` gets used 
        instead. Default to None
    log_like_func: Callable[[Any], Number], optional
        the log likelihood function :math:`\log(p(\mathbf{d}_{obs} \mid \mathbf{m}))`.
        It takes in a model :math:`\mathbf{m}` (any type is allowed, as long as it is
        consistent with the other arguments of this class) and returns the log
        of the likelihood function. This function is only used when ``log_like_ratio_func``
        is None. Default to None
    n_chains: int, optional
        the number of chains in the McMC sampling. Default is 10
    n_cpus: int, optional
        the number of CPUs available. If None (default) this is usually set to 
        the number of chains (``n_chains``)
    """

    def __init__(
        self,
        walkers_starting_states: List[Any],
        perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]],
        log_like_ratio_func: Union[LogLikelihood, Callable[[Any, Any], Number]] = None,
        log_like_func: Callable[[Any], Number] = None,
        n_chains: int = 10,
        n_cpus: int = None,
        save_dpred: bool = True,
    ):
        assert len(walkers_starting_states) == n_chains, (
            "``walkers_starting_states`` doesn't match the number of chains: "
            f"{len(walkers_starting_states)} != {n_chains}"
        )
        self.walkers_starting_states = walkers_starting_states
        self.perturbation_funcs = [
            _preprocess_func(func) for func in perturbation_funcs
        ]
        if isinstance(log_like_ratio_func, LogLikelihood):
            self.log_likelihood = log_like_ratio_func
        else:
            self.log_likelihood = LogLikelihood(
                log_like_ratio_func=log_like_ratio_func, 
                log_like_func=log_like_func, 
            ) 
        self.n_chains = n_chains
        self.n_cpus = n_cpus if n_cpus is not None else n_chains
        self.save_dpred = save_dpred
        self._chains = [
            BaseMarkovChain(
                id=i,
                starting_state=self.walkers_starting_states[i],
                perturbation_funcs=self.perturbation_funcs,
                log_likelihood=self.log_likelihood, 
                save_dpred=self.save_dpred,
            )
            for i in range(n_chains)
        ]
        
        self._init_repr_args()

    @property
    def chains(self) -> List[BaseMarkovChain]:
        """The ``MarkovChain`` instances of the current Bayesian inversion"""
        return self._chains
    
    @chains.setter
    def chains(self, updated_chains: List[BaseMarkovChain]):
        """setters of chains attached to this BayesianInversion instance

        Parameters
        ----------
        updated_chains : List[BaseMarkovChain]
            a list of chains to be set to current inversion

        Raises
        ------
        TypeError
            when ``updated_chains` is not a list or the elements are not instances of
            :class:`BaseMarkovChain`
        """
        if not isinstance(updated_chains, list) or \
            all([isinstance(c, BaseMarkovChain) for c in updated_chains]):
                raise TypeError("make sure the `updated_chains` is a list of chains")
        self._chains = updated_chains
        self.n_chains = len(updated_chains)

    def run(
        self,
        sampler: Sampler = None,
        n_iterations: int = 1000,
        burnin_iterations: int = 0,
        save_every: int = 100,
        verbose: bool = True,
        print_every: int = 100,
    ):
        r"""To run the inversion

        Parameters
        ----------
        sampler : bayesbay.samplers.Sampler, optional
            a sampler instance describing how chains intereact or modify their
            properties during sampling. This could be a sampler from the
            module :mod:`bayesbay.samplers` such as
            :class:`bayesbay.samplers.VanillaSampler` (default),
            :class:`bayesbay.samplers.ParallelTempering`, or
            :class:`bayesbay.samplers.SimulatedAnnealing`,
            or a customised sampler instance.
        n_iterations : int, optional
            total number of iterations to run, by default 1000
        burnin_iterations : int, optional
            the iteration number from which we start to save samples, by default 0
        save_every : int, optional
            the frequency with which we save the samples. By default a state is
            saved every after 100 iterations after the burn-in phase
        verbose : bool, optional
            whether to print the progress during sampling or not, by default True
        print_every : int, optional
            the frequency with which we print the progress and information during the
            sampling, by default 100 iterations
        """
        if sampler is None:
            sampler = VanillaSampler()
        sampler.initialize(self.chains)
        self._chains = sampler.run(
            n_iterations=n_iterations,
            n_cpus=self.n_cpus,
            burnin_iterations=burnin_iterations,
            save_every=save_every,
            verbose=verbose,
            print_every=print_every,
        )
    
    def get_results(
        self, 
        keys: Union[str, List[str]] = None,
        concatenate_chains: bool = True,
    ) -> Union[Dict[str, list], list]:
        """To get the saved states from current inversion

        Parameters
        ----------
        keys : Union[str, List[str]]
            key(s) to retrieve from the saved states. This will be ignored when states 
            are not of type :class:`State` or dict
        concatenate_chains : bool, optional
            whether to aggregate samples from all the Markov chains or to keep them
            seperate, by default True

        Returns
        -------
        Union[Dict[str, list], list]
            a dictionary from name of the attribute stored to the values, or a list of
            saved states (if the base level API is used, and states are not of type 
            :class:`State` or dict)
        """
        return self.get_results_from_chains(self.chains, keys, concatenate_chains)

    @staticmethod
    def get_results_from_chains(
        chains: Union[BaseMarkovChain, List[BaseMarkovChain]], 
        keys: Union[str, List[str]] = None,
        concatenate_chains: bool = True,
    ) -> Union[Dict[str, list], list]:
        """To get the saved states from a list of given Markov chains

        Parameters
        ----------
        chains : Union[BaseMarkovChain, List[BaseMarkovChain]]
            Markov chain(s) that the results are going to be extracted from
        keys : Union[str, List[str]]
            key(s) to retrieve from the saved states. This will be ignored when states 
            are not of type :class:`State` or dict
        concatenate_chains : bool, optional
            whether to aggregate samples from all the Markov chains or to keep them
            seperate, by default True

        Returns
        -------
        Union[Dict[str, list], list]
            a dictionary from name of the attribute stored to the values, or a list of
            saved states (if the base level API is used, and states are not of type 
            :class:`State` or dict)
        """
        if not isinstance(chains, list):
            chains = [chains]
        if not all([isinstance(c, BaseMarkovChain) for c in chains]):
            raise TypeError(
                "`chains` should be a list of Markov chains (i.e. instances of "
                "BaseMarkovChain or MarkovChain)"
            )
        if isinstance(keys, str):
            keys = [keys]
        if hasattr(chains[0].saved_states, "items"):
            results = defaultdict(list)
            for chain in chains:
                for key, saved_values in chain.saved_states.items():
                    if keys is None or key in keys:
                        if concatenate_chains and isinstance(saved_values, list):
                            results[key].extend(saved_values)
                        else:
                            results[key].append(saved_values)
        else:
            results = []
            for chain in chains:
                if concatenate_chains:
                    results.extend(chain.saved_states)
                else:
                    results.append(chain.saved_states)
        return dict(results)
    
    def _init_repr_args(self):
        self._repr_args = {
            "walkers_starting_states": self.walkers_starting_states, 
            "perturbation_funcs": self.perturbation_funcs, 
            "log_likelihood": self.log_likelihood, 
            "n_chains": self.n_chains, 
            "n_cpus": self.n_cpus, 
            "save_dpred": self.save_dpred, 
            "chains": self.chains, 
        }
    
    def __repr__(self) -> str:
        string = f"{self.__class__.__name__}("
        for k, v in self._repr_args.items():
            if hasattr(v, '__class__') and v.__class__.__repr__ is object.__repr__:
                repr_v = v.__class__.__name__
            elif k == "walkers_starting_states":
                repr_v = f"[{len(v)} arrays with shapes {', '.join(str(arr.shape) for arr in v)}]"
            elif k == "chains" and len(v) > 3: 
                repr_v = f"{str(v[:3])[:-1]}, ...{len(v)} chains in total...]"
            else:
                repr_v = repr(v)
            string += f"{k}={repr_v}, "
        return f"{string[:-2]})"
    
    def __str__(self) -> str:
        repr_args_copy = self._repr_args.copy() 
        _n_chains = ""
        if 'chains' in repr_args_copy and len(repr_args_copy['chains']) > 3:
            _n_chains = f"...{len(repr_args_copy['chains'])} chains in total..."
            repr_args_copy['chains'] = repr_args_copy['chains'][:3] + [_n_chains]
        string = f"{self.__class__.__name__}("
        string += pformat(repr_args_copy)[1:-1] + ")"
        string = string.replace(repr(_n_chains), _n_chains)     # remove the quotes
        return string
    

class BayesianInversion(BaseBayesianInversion):
    """A high-level class for performing Bayesian inversion using Markov Chain Monte
    Carlo (McMC) methods.

    This is a subclass of :class:`BaseBayesianInversion`.

    This class provides the basic structure for setting up and running McMC sampling,
    given user-configured parameterization settings, data targets and corresponding
    forward functions.

    Parameters
    ----------
    parameterization : bayesbay.parameterization.Parameterization
        pre-configured parameterization. This includes information about the dimension,
        parameterization bounds and properties of unknown parameterizations
    targets : List[bayesbay.Target]
        a list of data targets
    fwd_functions : Callable[[bayesbay.State], np.ndarray]
        a lsit of forward functions corresponding to each data targets provided above.
        Each function takes in a state and returns a numpy array of data predictions.
    n_chains: int, 10 by default
        the number of chains in the McMC sampling
    n_cpus: int, 10 by default
        the number of CPUs available. If None (default) this is usually set to 
        the number of chains (``n_chains``)
    """

    def __init__(
        self,
        parameterization: Parameterization,
        log_likelihood: LogLikelihood, 
        n_chains: int = 10,
        n_cpus: int = None,
        save_dpred: bool = True,
    ):
        if not isinstance(log_likelihood, LogLikelihood):
            raise TypeError(
                "``log_likelihood`` should be an instance of "
                "``bayesbay.LogLikelihood``"
            )

        self.parameterization = parameterization
        self.log_likelihood = log_likelihood
        self.n_chains = n_chains
        self.n_cpus = n_cpus if n_cpus is not None else n_chains
        self.save_dpred = save_dpred
        self.perturbation_funcs = self._init_perturbation_funcs()
        self._chains = [
            MarkovChain(
                id=i,
                parameterization=self.parameterization,
                log_likelihood=self.log_likelihood, 
                perturbation_funcs=self.perturbation_funcs, 
                saved_dpred=self.save_dpred,
            )
            for i in range(n_chains)
        ]
        self.log_likelihood.add_targets_observer(self)
        self._init_repr_args()
    
    def _init_repr_args(self) -> dict:
        self._repr_args = {
            "parameterization": self.parameterization, 
            "log_likelihood": self.log_likelihood, 
            "n_chains": self.n_chains, 
            "n_cpus": self.n_cpus, 
            "save_dpred": self.save_dpred, 
            "chains": self.chains, 
        }
        _parameterization = dict()
        for ps_name, ps in self.parameterization.parameter_spaces.items():
            _parameterization[ps_name] = {
                k: v for k, v in ps._repr_args.items() \
                    if k not in {"parameters", "name"}
            }
            _parameterization[ps_name]["parameters"] = list(ps.parameters.values())
        self._repr_args["parameterization"] = _parameterization

    def _init_perturbation_funcs(self) -> list:
        funcs_from_parameterization = self.parameterization.perturbation_functions
        funcs_from_log_likelihood = self.log_likelihood.perturbation_functions
        return funcs_from_parameterization + funcs_from_log_likelihood

    def update_log_likelihood_targets(self, targets: List[Target]):
        """function to be called by ``self.log_likelihood`` when there are more 
        target(s) added to the inversion. This method updates the perturbation function 
        list for the current inversion and its chains, and initializes the 
        ``current_state`` for each chain when there are hierarchical targets added.
        
        This method is called because we register ``self`` (i.e. the current 
        ``BayesianInversion``) as an observer to its log likelihood instance.

        Parameters
        ----------
        targets : List[Target]
            the list of targets added to ``self.log_likelihood``
        """
        self.perturbation_funcs = self._init_perturbation_funcs()
        for chain in self.chains:
            chain.update_perturbation_funcs(self.perturbation_funcs)
            chain.update_targets(targets)



from .exceptions import UserFunctionException
from .perturbations._base_perturbation import Perturbation


def _preprocess_func(func):
    if func is None:
        return None
    elif isinstance(func, Perturbation):
        return func
    f = None
    args = []
    kwargs = {}
    if isinstance(func, (tuple, list)) and len(func) > 1:
        f = func[0]
        if isinstance(func[1], (tuple, list)):
            args = func[1]
            if len(func) > 2 and isinstance(func[2], dict):
                kwargs = func[2]
        elif isinstance(func[1], dict):
            kwargs = func[1]
        else:
            raise TypeError(
                "additional arguments should be a list (i.e. args) or a dict (i.e. "
                f"kwargs), but got {func[1]} for the function instead"
            )
    elif isinstance(func, (tuple, list)):
        f = func[0]
    else:
        f = func
    return _FunctionWrapper(f, args, kwargs)


class _FunctionWrapper:
    """Function wrapper to make it pickleable (credit to emcee)"""

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = args or []
        self.kwargs = kwargs or {}

    def __call__(self, *args):
        try:
            return self.f(*args, *self.args, **self.kwargs)
        except Exception as e:
            raise UserFunctionException(e)

    @property
    def __name__(self) -> str:
        return self.f.__name__

    def __repr__(self) -> str:
        return self.__name__



from numbers import Number
from typing import Tuple, List
import random

from .._state import State
from ..parameters._parameters import Parameter
from ._base_perturbation import Perturbation


class ParamPerturbation(Perturbation):
    """Perturbation on a chosen parameter value

    Parameters
    ----------
    param_name : str
        the name of the parameter to be perturbed
    parameters : List[Parameter]
        list containing the :class:`Parameter` instances to be perturbed
    """

    def __init__(
        self,
        param_space_name: str,
        parameters: List[Parameter],
    ):
        self.param_space_name = param_space_name
        self.parameters = parameters

    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""perturb one value for each parameter in :attr:`self.parameters`, returning 
        a proposed state and the log of the corresponding partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        Parameters
        ----------
        state : State
            the current state to perturb from

        Returns
        -------
        Tuple[State, Number]
            proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        # randomly choose a position to perturb the value(s)
        old_ps_state = state[self.param_space_name]
        n_dims = old_ps_state.n_dimensions
        idx = random.randint(0, n_dims - 1)
        # randomly perturb the value(s)
        new_param_values = dict()
        log_prob_ratio = 0
        for param in self.parameters:
            if hasattr(param, "birth"):  # if it's a discretization
                new_ps_state, log_prob_ratio = param.perturb_value(old_ps_state, idx)
                new_state = state.copy()
                new_state.set_param_values(self.param_space_name, new_ps_state)
                return new_state, log_prob_ratio
            else:
                old_values = old_ps_state[param.name]
                new_param_values[param.name] = old_values.copy()
                old_value = old_values[idx]
                old_pos = old_ps_state["discretization"]
                pos = old_pos[idx] if old_pos is not None else None
                new_value, _ratio = param.perturb_value(old_value, pos)
                log_prob_ratio += _ratio
                new_param_values[param.name][idx] = new_value
        # structure new param value(s) into new state
        new_state = state.copy()
        new_state[self.param_space_name].param_values.update(new_param_values)
        return new_state, log_prob_ratio

    @property
    def __name__(self) -> str:
        param_names = [
            (f"{p.name}.discretization" if hasattr(p, "trans_d") else p.name) \
                for p in self.parameters
        ]
        param_names_str = (
            str(param_names) if len(param_names) != 1 else str(param_names[0])
        )
        return f"{self.type}({param_names_str})"



from typing import Union, List, Callable, Tuple, Any, Dict
from numbers import Number
from collections import defaultdict
import random
import math

from ._log_likelihood import LogLikelihood
from .parameterization import Parameterization
from ._state import State
from ._target import Target
from .exceptions import DimensionalityException, ForwardException, UserFunctionException


class BaseMarkovChain:
    r"""
    Low-level interface for a Markov Chain.

    Instantiation of this class is usually done by :class:`BaseBayesianInversion`.

    Parameters
    ----------
    id : Union[int, str]
        an integer or a string representing the ID of the current chain. For display
        purposes only
    starting_state : Any
        starting state of the current chain
    perturbation_funcs : List[Callable[[Any], Tuple[Any, Number]]]
        a list of perturbation functions
    log_likelihood : bayesbay.LogLikelihood
        instance of the ``bayesbay.LogLikelihood`` class
    temperature : int, optional
        used to temper the log likelihood, by default 1
    """

    def __init__(
        self,
        id: Union[int, str],
        starting_state: Any,
        perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]],
        log_likelihood: LogLikelihood,
        temperature: float = 1,
        save_dpred: bool = True,
    ):
        self.id = id
        self.current_state = starting_state
        self._temperature = temperature
        self.log_likelihood = log_likelihood
        self.save_dpred = save_dpred
        self.update_perturbation_funcs(perturbation_funcs)
        self._init_statistics()
        self._init_saved_states()

    @property
    def current_state(self) -> Union[State, Any]:
        return self._current_state
    
    @current_state.setter
    def current_state(self, state: Union[State, Any]):
        self._current_state = state

    @property
    def temperature(self) -> Number:
        """The current temperature used in the log likelihood ratio"""
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        self._temperature = value
        if isinstance(self.current_state, State):
            self.current_state.temperature = value

    @property
    def saved_states(self) -> Union[Dict[str, list], list]:
        """All the states saved so far"""
        return getattr(self, "_saved_states", None)

    @property
    def statistics(self):
        return self._statistics
    
    @property
    def ith_iteration(self):
        return self.statistics["n_proposed_models_total"]

    def update_perturbation_funcs(self, funcs: List[Callable]):
        self.perturbation_funcs = funcs
        self.perturbation_types = [func.__name__ for func in funcs]
    
    def update_targets(self, targets: List[Target]):
        for target in targets:
            target.initialize(self.current_state)
    
    def _init_saved_states(self):
        if isinstance(self.current_state, (State, dict)):
            self._saved_states = defaultdict(list)
        else:
            self._saved_states = []

    def _init_statistics(self):
        self._statistics = {
            "n_proposed_models": defaultdict(int),
            "n_accepted_models": defaultdict(int),
            "n_proposed_models_total": 0,
            "n_accepted_models_total": 0,
            "exceptions": defaultdict(int),
        }

    def _save_state(self):
        if isinstance(self.current_state, (State, dict)):
            for k, v in self.current_state.items():
                if (
                    "n_dimensions" not in k
                    or (
                        hasattr(self, "parameterization") and
                        self.parameterization.parameter_spaces[
                        k[: -len(".n_dimensions")]
                    ].trans_d)
                ):
                    self.saved_states[k].append(v)
            if self.save_dpred:
                for k, v in self.current_state.cache.items():
                    if "dpred" in k:
                        self.saved_states[k].append(v)
        else:
            self.saved_states.append(self.current_state)

    def _save_statistics(self, perturb_i, accepted):
        perturb_type = self.perturbation_types[perturb_i]
        self._statistics["n_proposed_models"][perturb_type] += 1
        self._statistics["n_accepted_models"][perturb_type] += 1 if accepted else 0
        self._statistics["n_proposed_models_total"] += 1
        self._statistics["n_accepted_models_total"] += 1 if accepted else 0

    def print_statistics(self):
        """print the statistics about the Markov Chain history, including the number of
        explored and accepted states for each perturbation, acceptance rates and the 
        current temperature
        """
        head = f"Chain ID: {self.id}"
        head += f"\nTEMPERATURE: {self.temperature}"
        head += f"\nEXPLORED MODELS: {self.statistics['n_proposed_models_total']}"
        _accepted_total = self.statistics["n_accepted_models_total"]
        _explored_total = self.statistics["n_proposed_models_total"]
        acceptance_rate = _accepted_total / _explored_total * 100
        head += "\nACCEPTANCE RATE: %d/%d (%.2f %%)" % (
            _accepted_total,
            _explored_total,
            acceptance_rate,
        )
        print(head)
        print("PARTIAL ACCEPTANCE RATES:")
        _accepted_all = self.statistics["n_accepted_models"]
        _explored_all = self.statistics["n_proposed_models"]
        for perturb_type in sorted(self.statistics["n_proposed_models"]):
            _explored = _explored_all[perturb_type]
            _accepted = _accepted_all[perturb_type]
            acceptance_rate = _accepted / _explored * 100
            print(
                "\t%s: %d/%d (%.2f%%)"
                % (perturb_type, _accepted, _explored, acceptance_rate)
            )

    def _log_likelihood_ratio(self, new_state):
        return self.log_likelihood.log_likelihood_ratio(
            self.current_state, new_state, self.temperature
        )

    def _next_iteration(self):
        _last_exception = None
        for i in range(500):
            # choose one perturbation function and type
            i_perturb = random.randint(0, len(self.perturbation_funcs) - 1)
            perturb_func = self.perturbation_funcs[i_perturb]

            # perturb and get the partial acceptance probability excluding log
            # likelihood ratio
            try:
                new_state, log_prob_ratio = perturb_func(self.current_state)
            except (DimensionalityException, UserFunctionException) as e:
                _last_exception = e
                self._statistics["exceptions"][e.__class__.__name__] += 1
                continue

            # calculate the log likelihood ratio
            try:
                log_likelihood_ratio = self._log_likelihood_ratio(new_state)
            except (ForwardException, UserFunctionException) as e:
                _last_exception = e
                self._statistics["exceptions"][e.__class__.__name__] += 1
                continue

            # decide whether to accept
            acceptance_probability = log_prob_ratio + log_likelihood_ratio
            accepted = acceptance_probability > math.log(random.random())
            if accepted:
                self.current_state = new_state

            # save statistics and current state
            self._save_statistics(i_perturb, accepted)
            if self.save_current_iteration and self.temperature == 1.0:
                self._save_state()
            return
        raise RuntimeError(
            f"Chain {self.id} failed in perturb or forward calculation for 500 times. "
            "See above for the last exception."
        ) from _last_exception

    def advance_chain(
        self,
        n_iterations: int = 1000,
        burnin_iterations: int = 0,
        save_every: int = 100,
        verbose: bool = True,
        print_every: int = 100,
        begin_iteration: Callable[["BaseMarkovChain"], None] = None,
        end_iteration: Callable[["BaseMarkovChain"], None] = None,
    ):
        """advance the chain for a given number of iterations

        Parameters
        ----------
        n_iterations : int, optional
            the number of iterations to advance, by default 1000
        burnin_iterations : int, optional
            the iteration number from which we start to save samples, by default 0
        save_every : int, optional
            the frequency in which we save the samples, by default 100
        verbose : bool, optional
            whether to print the progress during sampling or not, by default True
        print_every : int, optional
            the frequency with which we print the progress and information during the
            sampling, by default 100 iterations
        begin_iteration : Callable[["BaseMarkovChain"], None], optional
            customized function that's to be run at before an iteration
        end_iteration : Callable[["BaseMarkovChain"], None], optional
            customized function that's to be run at after an iteration

        Returns
        -------
        BaseMarkovChain
            the chain itself is returned
        """
        for _ in range(n_iterations):
            i = self.ith_iteration + 1
            if i <= burnin_iterations:
                self.save_current_iteration = False
            else:
                self.save_current_iteration = not (i - burnin_iterations) % save_every

            begin_iteration(self)
            self._next_iteration()
            end_iteration(self)
            if verbose and not i % print_every:
                self.print_statistics()

        return self

    def _repr_args(self) -> dict:
        return {
            "id": self.id, 
            "temperature": self.temperature, 
            "n_proposed_models_total": self.statistics["n_proposed_models_total"], 
            "n_accepted_models_total": self.statistics["n_accepted_models_total"], 
        }
    
    def __repr__(self) -> str:
        string = f"{self.__class__.__name__}("
        for k, v in self._repr_args().items():
            string += f"{k}={v}, "
        return f"{string[:-2]})"


class MarkovChain(BaseMarkovChain):
    """High-level interface for a Markov Chain.

    This is a subclass of :class:`BaseMarkovChain`.

    Instantiation of this class is usually done by :class:`BayesianInversion`.

    Parameters
    ----------
    id : Union[int, str]
        an integer or a string representing the ID of the current chain. For display
        purposes only
    parameterization : bayesbay.parameterization.Parameterization
        pre-configured parameterization. This includes information about the dimension,
        parameterization bounds and properties of unknown parameterizations
    log_likelihood : bayesbay.LogLikelihood
        instance of the ``bayesbay.LogLikelihood`` class
    perturbation_funcs : List[Callable]
        a list of perturbation functions (generated automatically by 
        :class:`BayesianInversion` from ``parameterization`` and ``log_likelihood``)
    temperature : int, optional
        used to temper the log likelihood, by default 1
    """

    def __init__(
        self,
        id: Union[int, str],
        parameterization: Parameterization,
        log_likelihood: LogLikelihood, 
        perturbation_funcs: List[Callable], 
        temperature: float = 1,
        saved_dpred: bool = True,
    ):
        self.parameterization = parameterization
        self.log_likelihood = log_likelihood
        starting_state = self._init_starting_state()
        super().__init__(
            id=id, 
            starting_state=starting_state, 
            perturbation_funcs=perturbation_funcs, 
            log_likelihood=log_likelihood, 
            temperature=temperature,
            save_dpred=saved_dpred, 
        )

    def _init_starting_state(self) -> State:
        """Initialize the parameterization by defining a starting state."""
        starting_state = self.parameterization.initialize()
        self.log_likelihood.initialize(starting_state)
        return starting_state
    


from numbers import Number
from typing import Tuple, List
import random

from .._state import State, DataNoiseState
from .._target import Target
from ._base_perturbation import Perturbation


class NoisePerturbation(Perturbation):
    """Perturbation by changing the noise level estimation

    Parameters
    ----------
    targets : List[Target]
    """
    def __init__(
        self,
        targets: List[Target]
    ):
        self.targets = targets

    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""proposes a new state by perturbing the data noise of the given
        state and calculates the log of the corresponding partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},
        
        which in this case equals zero since a uniform probability distribution is
        inherently assumed for the noise properties.

        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            the proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert) = 0`
        """
        new_data_noise_all = dict()
        for target in self.targets:
            old_data_noise_state = state[target.name]
            new_data_noise_state = self._perturb_target(target, old_data_noise_state)
            new_data_noise_all[target.name] = new_data_noise_state
        new_state = state.copy(keep_dpred=True)
        new_state.param_values.update(new_data_noise_all)
        return new_state, 0
    
    def _perturb_target(
        self, target: Target, data_noise_state: DataNoiseState
    ) -> DataNoiseState:
        to_be_perturbed = ["std"]
        if target.noise_is_correlated:
            to_be_perturbed.append("correlation")
        new_values = {"correlation": None}
        for p in to_be_perturbed:
            vmin = getattr(target, f"{p}_min")
            vmax = getattr(target, f"{p}_max")
            perturb_std = getattr(target, f"{p}_perturb_std")
            old_value = getattr(data_noise_state, p)
            while True:
                random_deviate = random.normalvariate(0, perturb_std)
                new_value = old_value + random_deviate
                if new_value < vmin or new_value > vmax:
                    continue
                break
            new_values[p] = new_value
        return DataNoiseState(**new_values)

    @property
    def __name__(self) -> str:
        """Identifier for the perturbation"""
        target_names = [t.name for t in self.targets]
        target_names_str = (
            str(target_names) if len(target_names) > 1 else str(target_names[0])
        )
        return f"{self.type}({target_names_str})"



from ._parameters import (
    Parameter,
    UniformParameter,
    GaussianParameter,
    CustomParameter,
)


__all__ = [
    "Parameter",
    "UniformParameter",
    "GaussianParameter",
    "CustomParameter",
]



from typing import List, Callable, Tuple, Dict
from numbers import Number
import random
import numpy as np

from .._state import State, ParameterSpaceState
from ..exceptions import DimensionalityException
from ..parameters import Parameter
from ..perturbations._param_values import ParamPerturbation
from ..perturbations._birth_death import BirthPerturbation, DeathPerturbation
from .._utils_1d import delete, insert_scalar


class ParameterSpace:
    r"""Utility class to parameterize the Bayesian inference problem
    
    Parameters
    ----------
    name : str
        name of the parameter space, for display and storing purposes
    n_dimensions : Number, optional
        number of dimensions defining the parameter space. None (default) results 
        in a trans-dimensional space.
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        parameter space is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range ``n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if 
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and 
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is
        
        .. code-block:: python
            
            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)
            
    parameters : List[Parameter], optional
        a list of free parameters, by default None
    """
    def __init__(
        self, 
        name: str,
        n_dimensions: int = None, 
        n_dimensions_min: int = 1, 
        n_dimensions_max: int = 10, 
        n_dimensions_init_range: Number = 0.3, 
        parameters: List[Parameter] = None, 
    ):
        self._name = name
        self._trans_d = n_dimensions is None
        self._n_dimensions = n_dimensions
        self._n_dimensions_min = n_dimensions_min if self._trans_d else n_dimensions
        self._n_dimensions_max = n_dimensions_max if self._trans_d else n_dimensions
        self._n_dimensions_init_range = n_dimensions_init_range
        self._parameters = dict()
        if parameters is not None:
            for param in parameters:
                self._parameters[param.name] = param
        self._init_perturbation_funcs()
        self._init_repr_args()
    
    @property
    def name(self) -> str:
        """name of the parameter space"""
        return self._name
    
    @property
    def trans_d(self) -> bool:
        """indicates whether the current configuration allows changes in 
        dimensionality
        """
        return self._trans_d
    
    @property
    def parameters(self) -> Dict[str, Parameter]:
        """all the free parameters defined in this parameter space"""
        return self._parameters
    
    @property
    def perturbation_functions(self) -> List[Callable[[State], Tuple[State, Number]]]:
        r"""a list of perturbation functions allowed in the current parameter space.
        Each function takes in a state (see :class:`State`) and returns a new
        state along with the corresponding partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},
            
        """
        return self._perturbation_funcs
    
    def initialize(self) -> ParameterSpaceState:
        """initializes the parameter space including its parameter values

        Returns
        -------
        ParameterSpaceState
            an initial parameter space state
        """
        # initialize number of dimensions
        if not self.trans_d:
            n_dimensions = self._n_dimensions
        else:
            init_range = self._n_dimensions_init_range
            n_dims_min = self._n_dimensions_min
            n_dims_max = self._n_dimensions_max
            init_max = int((n_dims_max - n_dims_min) * init_range + n_dims_min)
            n_dimensions = random.randint(n_dims_min, init_max)
        # initialize parameter values
        parameter_vals = dict()
        for name, param in self.parameters.items():
            parameter_vals[name] = param.initialize(np.empty(n_dimensions))
        return ParameterSpaceState(n_dimensions, parameter_vals)
    
    def log_prior(self, param_space_state: ParameterSpaceState) -> Number:
        joint_log_prior = 0
        for param_name, param in self.parameters.items():
            param_values = param_space_state[param_name]
            all_log_priors = [param.log_prior(v, None) for v in param_values]
            joint_log_prior += sum(all_log_priors)
        return joint_log_prior
    
    def birth(self, ps_state: ParameterSpaceState) -> Tuple[ParameterSpaceState, float]:
        r"""adds a dimension to the current parameter space and returns the
        thus obtained new state along with the log of the corresponding partial 
        acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
            
        In this case, we have
        
        .. math::
            \frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)} = \prod_i{p\left({m_i^{k+1}}\right)},
                         
        because all entries of the :math:`i`\ th free parameter, :math:`\mathbf{m}_i`, 
        are equal to those of :math:`\mathbf{m'}_i` except for the newly born 
        :math:`m_i^{k+1}`, with :math:`k` denoting the number of dimensions in 
        the parameter space prior to the birth perturbation.
        
        Furthermore, since :math:`m_i^{k+1}` is drawn from the prior and there are
        :math:`k+1` positions available to randomly remove a dimension or add a 
        new one,
        
        .. math::
            \frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)} =
            \frac{1}{k+1} \frac{\prod_i{p\left({m_i^{k+1}}\right)}}{\frac{1}{(k+1)}} =
            \frac{1}{\prod_i{p\left({m_i^{k+1}}\right)}}.
            
        
        Finally, it is easy to shown that in this case :math:`\lvert \mathbf{J} \rvert = 1`. 
        It follows that
                
        .. math::
            \alpha_{p} = 
            \frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}
            \frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            = \prod_i{p\left({\mathbf{m}_i^{k+1}}\right)} \frac{1}{\prod_i{p\left({m_i^{k+1}}\right)}}
            = 1,
        
        and :math:`\log(\alpha_{p}) = 0`.
        
        Parameters
        ----------
        ParameterSpaceState
            initial parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert) = 0`
        """
        n_dims = ps_state.n_dimensions
        if n_dims == self._n_dimensions_max:
            raise DimensionalityException("Birth")
        i_insert = random.randint(0, n_dims)
        new_param_values = dict()
        for param_name, param_vals in ps_state.param_values.items():
            new_param_values[param_name] = insert_scalar(
                param_vals, 
                i_insert, 
                self.parameters[param_name].sample()
            )
        new_state = ParameterSpaceState(n_dims+1, new_param_values)
        prob_ratio = 0
        return new_state, prob_ratio
    
    def death(self, ps_state: ParameterSpaceState) -> Tuple[ParameterSpaceState, float]:
        r"""removes a dimension from the given parameter space and returns the
        thus obtained new state along with the log of the corresponding partial 
        acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
            
        Following a reasoning similar to that explained in the documentation of
        :meth:`birth`, in this case :math:`\log(\alpha_{p}) = 0`.
        
        Parameters
        ----------
        ParameterSpaceState
            initial parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert) = 0`
        """
        n_dims = ps_state.n_dimensions
        if n_dims == self._n_dimensions_min:
            raise DimensionalityException("Death")
        i_to_remove = random.randint(0, n_dims-1)
        new_param_values = dict()
        for param_name, param_vals in ps_state.param_values.items():
            new_param_values[param_name] = delete(param_vals, i_to_remove)
        new_state = ParameterSpaceState(n_dims-1, new_param_values)
        prob_ratio = 0
        return new_state, prob_ratio
    
    def _init_perturbation_funcs(self):
        self._perturbation_funcs = []
        if self.parameters:
            self._perturbation_funcs.append(
                ParamPerturbation(self.name, list(self.parameters.values()))
            )
        if self.trans_d:
            self._perturbation_funcs.append(BirthPerturbation(self))
            self._perturbation_funcs.append(DeathPerturbation(self))

    def _init_repr_args(self):
        self._repr_args = {
            "name": self.name, "parameters": list(self.parameters.values())
        }
        if self.trans_d:
            self._repr_args["n_dimensions_min"] = self._n_dimensions_min
            self._repr_args["n_dimensions_max"] = self._n_dimensions_max
            self._repr_args["n_dimensions_init_range"] = self._n_dimensions_init_range
        else:
            self._repr_args["n_dimensions"] = self._n_dimensions

    def __repr__(self) -> str:
        attr_to_show = self._repr_args
        string = f"{attr_to_show['name']}("
        for k, v in attr_to_show.items():
            string += f"{k}={v}, " if k != "name" else ""
        return f"{string[:-2]})"



import math
from typing import Any, List, Callable, Tuple, Union
from numbers import Number
import numpy as np

from .exceptions import ForwardException
from ._state import State
from ._target import Target
from .perturbations._data_noise import NoisePerturbation
from ._utils import _preprocess_func


class LogLikelihood:
    r"""Helper class to evaluate the log likelihood ratio

    One out of the following three sets of input needs to be supplied:

    1. ``target`` and ``fwd_functions``
    2. ``log_like_ratio_func``
    3. ``log_like_func``

    The input will be taken in precedence as above. For example, if both ``target``
    and ``fwd_functions`` are supplied, then ``log_like_func`` and
    ``log_like_ratio_func`` are ignored regardless of their values, and so on.

    This class then will wrap the input functions properly so that the
    :meth:`log_likelihood_ratio` is ready for use by the Markov chains.

    If ``targets`` are supplied and if any of them have unknown data noise, and if the
    high level API is used (i.e. ``BayesianInversion`` instead of
    ``BaseBayesianInversion`` is used), then the perturbations of the data noises will
    be taken into account regardless of whether ``fwd_functions`` is given, and the
    state for each chain will be initialized accordingly for the unknown data noise.

    Parameters
    ----------
    targets : bayesbay.Target, optional
        a list of data targets, default to None
    fwd_functions : List[Callable[[bayesbay.State], np.ndarray]], optional
        a list of forward functions corresponding to each data targets provided above.
        Each function takes in a state and produces a numpy array of data predictions.
        Default to None.
    log_like_ratio_func: Callable[[Any, Any], Number], optional
        a function to calculate the log likelihood ratio 
        :math:`\log(\frac{p(\mathbf{d}_{obs} \mid \mathbf{m}')}{p(\mathbf{d}_{obs} \mid \mathbf{m})})`.
        It takes the current and proposed models, :math:`\mathbf{m}` and :math:`\mathbf{m'}`, 
        whose type should be consistent with the other arguments of this class, and returns a scalar. 
        This is utilised in the calculation of the acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},
        
        where :math:`\mathbf{J}` denotes the Jacobian of the transformation.        
        If None, ``log_like_func`` gets used instead. Default to None
    log_like_func: Callable[[Any], Number], optional
        a function to calculate the log likelihood 
        :math:`\log(p(\mathbf{d}_{obs} \mid \mathbf{m}))`.
        It takes in a model :math:`\mathbf{m}` (any type is allowed, as long as it is
        consistent with the other arguments of this class) and returns a
        scalar. This function is only used when ``log_like_ratio_func``
        is None. Default to None
    """

    def __init__(
        self,
        targets: Union[Target, List[Target]] = None,
        fwd_functions: Union[Callable, List[Callable[[State], np.ndarray]]] = None,
        log_like_ratio_func: Callable[[Any, Any], Number] = None,
        log_like_func: Callable[[Any], Number] = None,
    ):
        self._perturbation_funcs_observers = []
        self.add_targets(targets, fwd_functions)
        self.log_like_ratio_func = log_like_ratio_func
        self.log_like_func = log_like_func
        self._init_log_likelihood_ratio()
        
    @property
    def targets(self) -> List[Target]:
        """list of targets associated with the current log likelihood instance"""
        if not hasattr(self, "_targets"):
            self._targets = []
        return self._targets
    
    @property
    def fwd_functions(self) -> List[Callable]:
        """list of forward functions associated with the current log likelihood
        instance"""
        if not hasattr(self, "_fwd_functions"):
            self._fwd_functions = []
        return self._fwd_functions

    @property
    def perturbation_functions(self) -> List[Callable[[State], Tuple[State, Number]]]:
        """A list of perturbation functions associated with the data noise of the
        provided targets.

        Perturbation functions are included in this list only when the data noise of
        the target(s) is explicitly set to be unknown(s).
        """
        return self._perturbation_funcs

    def initialize(self, state: State):
        """initialize the starting state of data noise associated with the targets in 
        current log likelihood that are hierarchical (i.e. having unknown data noise)

        Parameters
        ----------
        state : State
            the state to be updated on
        """
        if self.targets is not None:
            for target in self.targets:
                target.initialize(state)
    
    def add_targets_observer(self, inversion):
        """add an observer (typically an instance of the high level 
        :class:`BayesianInversion`). When the data targets are extended with
        new observations, the observers will be notified by being called 
        ``update_log_likelihood_targets`` method

        Parameters
        ----------
        inversion : BayesianInversion
            the observer to be added to the notification list
        """
        self._perturbation_funcs_observers.append(inversion)
    
    def add_targets(
        self, 
        targets: Union[Target, List[Target]] = None, 
        fwd_functions: Union[Callable, List[Callable[[State], np.ndarray]]] = None
    ):
        """add new target(s) and its/their associated forward function(s)

        Parameters
        ----------
        targets : Union[Target, List[Target]], optional
            the targets to be added, by default None
        fwd_functions : Union[Callable, List[Callable[[State], np.ndarray]]], optional
            the forward functions to be added, by default None

        Raises
        ------
        TypeError
            when the ``targets`` isn't a list of Target of a single Target instance, 
            or when the ``fwd_functions`` isn't a list of functions or a single 
            function
        """
        if targets is not None:
            if not isinstance(targets, list):
                targets = [targets]
            for target in targets:
                if not isinstance(target, Target):
                    raise TypeError("`targets` should either be a list of Target instances or a Target")
            self.targets.extend(targets)
        
        if fwd_functions is not None:
            if not isinstance(fwd_functions, list):
                fwd_functions = [fwd_functions]
            for func in fwd_functions:
                if not isinstance(func, (Callable, tuple)):
                    raise TypeError("`fwd_functions` should be a Callable/tuple or a list of Callables/tuples")
            self.fwd_functions.extend(fwd_functions)
        
        self._check_duplicate_target_names()
        self._init_perturbation_funcs()
        for inversion in self._perturbation_funcs_observers:
            inversion.update_log_likelihood_targets(targets)

    def log_likelihood_ratio(
        self,
        old_state: Union[State, Any],
        new_state: Union[State, Any],
        temperature: Number = 1,
    ) -> Number:
        r"""Returns the (possibly tempered) log of the likelihood ratio

        .. math::
            \left[
                \frac{p\left(\mathbf{d}_{obs} \mid \mathbf{m'}\right)}{p\left(\mathbf{d}_{obs} \mid \mathbf{m}\right)}
            \right]^{\frac{1}{T}}
            =
            \left[
            \frac{\lvert \mathbf{C}_e \rvert}{\lvert \mathbf{C}^{\prime}_e \rvert}
            \exp\left(- \frac{\Phi(\mathbf{m'}) - \Phi(\mathbf{m})}{2}\right)
            \right]^{\frac{1}{T}},


        where :math:`\mathbf{C}_e` denotes the data covariance matrix,
        :math:`\Phi(\mathbf{m})` the data misfit associated with the model
        :math:`\mathbf{m}`, :math:`T` the chain temperature, and the prime
        superscript indicates that the model has been perturbed.


        Parameters
        ----------
        old_state : bayesbay.State
            the state of the Bayesian inference prior to the model perturbation
        new_state : bayesbay.State
            the state of the Bayesian inference after the model perturbation
        temperature : Number
            the temperature associated with current chain and iteration

        Returns
        -------
        Number
            log likelihood ratio
        """
        return self._log_likelihood_ratio(old_state, new_state) / temperature

    def _log_likelihood_ratio_from_targets(
        self, old_state: State, new_state: State
    ) -> Number:
        old_misfit, old_log_det = self._get_misfit_and_det(old_state)
        new_misfit, new_log_det = self._get_misfit_and_det(new_state)
        log_like_ratio = (old_log_det - new_log_det) + (old_misfit - new_misfit) / 2
        return log_like_ratio

    def _log_likelihood_ratio_from_loglike(
        self, old_state: Union[State, Any], new_state: Union[State, Any]
    ) -> Number:
        old_loglike = self.log_like_func(old_state)
        new_loglike = self.log_like_func(new_state)
        return new_loglike - old_loglike

    def _log_likelihood_ratio_from_loglike_ratio(
        self, old_state: Union[State, Any], new_state: Union[State, Any]
    ) -> Number:
        return self.log_like_ratio_func(old_state, new_state)

    def _check_duplicate_target_names(self):
        if self.targets is not None:
            all_target_names = [t.name for t in self.targets]
            if len(all_target_names) != len(set(all_target_names)):
                raise ValueError("duplicate target names found")

    def _init_log_likelihood_ratio(self):
        if self.targets and self.fwd_functions:
            _fwd_functions = self.fwd_functions
            _targets = self.targets
            assert len(_fwd_functions) == len(_targets)
            self._targets = _targets
            self._fwd_functions = [_preprocess_func(func) for func in _fwd_functions]
            self._log_likelihood_ratio = self._log_likelihood_ratio_from_targets
        elif self.log_like_ratio_func:
            self.log_like_ratio_func = _preprocess_func(self.log_like_ratio_func)
            self._log_likelihood_ratio = self._log_likelihood_ratio_from_loglike_ratio
        elif self.log_like_func:
            self.log_like_func = _preprocess_func(self.log_like_func)
            self._log_likelihood_ratio = self._log_likelihood_ratio_from_loglike
        else:
            raise ValueError(
                "please provide one out of the following three sets of input: \n"
                "\t1. ``targets`` and ``fwd_functions``\n"
                "\t2. ``log_like_ratio_func``\n"
                "\t3. ``log_like_func``"
            )

    def _init_perturbation_funcs(self) -> list:
        if self.targets is not None:
            hier_targets = (
                [t for t in self.targets if t.is_hierarchical]
                if self.targets is not None
                else []
            )
            self._perturbation_funcs = (
                [NoisePerturbation(hier_targets)] if hier_targets else []
            )
        else:
            self._perturbation_funcs = []

    def _get_misfit_and_det(self, state: State) -> Tuple[Number, Number]:
        misfit = 0
        log_det = 0
        for target, fwd_func in zip(self.targets, self.fwd_functions):
            _dpred_key = f"{target.name}.dpred"
            if state.saved_in_cache(_dpred_key):
                dpred = state.load_from_cache(_dpred_key)
            else:
                try:
                    dpred = fwd_func(state)
                except Exception as e:
                    raise ForwardException(e)
                state.save_to_cache(_dpred_key, dpred)
            residual = dpred - target.dobs
            misfit += residual @ target.inverse_covariance_times_vector(state, residual)
            if target.is_hierarchical:
                log_det += target.log_determinant_covariance(state)
        if not isinstance(misfit, Number):
            raise TypeError(
                f"misfit is expected to be of a Number type, but is {type(misfit)} "
                "instead. This might be due to predicted data not being 1D. Try using "
                "numpy.squeeze() function on predicted data before returning it from "
                "your forward function"
            )
        return misfit, log_det



from typing import Tuple
from numbers import Number

from ._base_perturbation import Perturbation
from .._state import State


class BirthPerturbation(Perturbation):
    """Perturbation by creating a new dimension
    
    Parameters
    ----------
    parameter_space : ParameterSpace
        instance of :class:`bayesbay.parameterization.ParameterSpace`
    """
    def __init__(self, parameter_space):
        self.param_space = parameter_space
        self.param_space_name = parameter_space.name

    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""proposes a new state by adding a dimension to the given parameter 
        space (:attr:`self.param_space`) and calculates the log of the corresponding
        partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},
        
        where :math:`{\bf d}_{obs}` denotes the observed data and 
        :math:`\mathbf{J}` the Jacobian of the transformation.            
        
        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            the proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`

        Raises
        ------
        DimensionalityException
            when the current state has already reached the maximum number of Voronoi
            cells
        """
        old_ps_state = state[self.param_space_name]
        new_ps_state, log_prob_ratio = self.param_space.birth(old_ps_state)
        new_state = state.copy()
        new_state.set_param_values(self.param_space_name, new_ps_state)
        return new_state, log_prob_ratio

    @property
    def __name__(self) -> str:
        return f"{self.type}({self.param_space_name})"


class DeathPerturbation(Perturbation):
    """Perturbation by removing an existing dimension
    
    Parameters
    ----------
    parameter_space : ParameterSpace
        instance of :class:`bayesbay.parameterization.ParameterSpace`
    """
    def __init__(self, parameter_space,):
        self.param_space = parameter_space
        self.param_space_name = parameter_space.name

    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""proposes a new state by removing a dimension from the given parameter 
        space (:attr:`self.param_space`) and calculates the log of the corresponding
        partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}
        
        where :math:`{\bf d}_{obs}` denotes the observed data and 
        :math:`\mathbf{J}` the Jacobian of the transformation.            
        
        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            the proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`

        Raises
        ------
        DimensionalityException
            when the current state has already reached the minimum number of Voronoi
            cells
        """
        old_ps_state = state[self.param_space_name]
        new_ps_state, log_prob_ratio = self.param_space.death(old_ps_state)
        new_state = state.copy()
        new_state.set_param_values(self.param_space_name, new_ps_state)
        return new_state, log_prob_ratio
 
    @property
    def __name__(self) -> str:
        return f"{self.type}({self.param_space_name})"



from ._base_perturbation import Perturbation
from ._data_noise import NoisePerturbation
from ._param_values import ParamPerturbation
from ._birth_death import (
    BirthPerturbation,
    DeathPerturbation,
)


__all__ = [
    "Perturbation",
    "NoisePerturbation",
    "ParamPerturbation",
    "BirthPerturbation",
    "DeathPerturbation",
]



from typing import List, Union
from pprint import pformat

from ._parameter_space import ParameterSpace
from ..perturbations import Perturbation
from .._state import State


class Parameterization:
    """Parameterization setting that consists of one or more :class:`ParameterSpace`
    instances
    
    Parameters
    ----------
    parameter_space : Union[ParameterSpace, List[ParameterSpace]]
        one or more :class:`ParameterSpace` instance(s)
    """
    def __init__(self, parameter_space: Union[ParameterSpace, List[ParameterSpace]]):
        parameter_spaces = parameter_space if isinstance(parameter_space, list) else [parameter_space]
        self._check_duplicate_param_space_names(parameter_spaces)
        self._check_duplicate_param_names(parameter_spaces)
        
        self.parameter_spaces = {ps.name: ps for ps in parameter_spaces}
        self._init_perturbation_funcs()
    
    def _check_duplicate_param_space_names(self, parameter_spaces: List[ParameterSpace]):
        all_param_space_names = [ps.name for ps in parameter_spaces]
        if len(all_param_space_names) != len(set(all_param_space_names)):
            raise ValueError("duplicate parameter space names found")
    
    def _check_duplicate_param_names(self, parameter_spaces: List[ParameterSpace]):
        all_param_names = []
        for ps in parameter_spaces:
            param_names = list(ps.parameters.keys())
            all_param_names.extend(param_names)
        if len(all_param_names) != len(set(all_param_names)):
            raise ValueError(
                "duplicate parameter names found, or one parameter has been assigned "
                "to different parameter spaces"
            )

    def _init_perturbation_funcs(self):
        self._perturbation_funcs = []
        for ps in self.parameter_spaces.values():
            self._perturbation_funcs.extend(ps.perturbation_functions)
    
    def initialize(self) -> State:
        """initializes the parameter space(s) constituting the parameterization
        
        Returns
        -------
        State
            Numerical values corresponding to the free parameters in the
            parameter space(s) constituting the parameterization
        """
        param_values = dict()
        for ps_name, ps in self.parameter_spaces.items():
            param_values[ps_name] = ps.initialize()
        return State(param_values)
    
    @property
    def perturbation_functions(self) -> List[Perturbation]:
        """the list of perturbation functions allowed to perturb the
        parameterization
        """
        return self._perturbation_funcs
    
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(parameter_spaces="
            f"{list(self.parameter_spaces.values())})"
        )
    
    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{pformat(list(self.parameter_spaces.values()))})"
        )



from typing import Union
from numbers import Number
import random
import math
import numpy as np

from ._utils_1d import inverse_covariance
from ._state import State, DataNoiseState


class Target:
    """Observed data with noise that can be treated as an unknown

    Parameters
    ----------
    name : str
        name of the data, for display purposes
    dobs : np.ndarray
        numerical data
    covariance_mat_inv : Union[Number, np.ndarray], optional
        the inverse of the data covariance matrix, either a number or a full matrix, by
        default None
    noise_is_correlated : bool, optional
        whether the noise between data points is correlated or not, by default False
    std_min : Number, optional
        the minimum value of the standard deviation of data noise, by default 0.01
    std_max : Number, optional
        the maximum value of the standard deviation of data noise, by default 1
    std_perturb_std : Number, optional
        the perturbation standard deviation of the standard deviation of data noise, by
        default 0.1
    correlation_min : Number, optional
        the miminum value of the correlation of data noise, by default 0.01
    correlation_max : Number, optional
        the maximum value of the correlation of data noise, by default 1
    correlation_perturb_std : Number, optional
        the perturbation standard deviation of the standard deviation of data noise, by
        default 0.1
    """

    def __init__(
        self,
        name: str,
        dobs: np.ndarray,
        covariance_mat_inv: Union[Number, np.ndarray] = None,
        noise_is_correlated: bool = False,
        std_min: Number = 0.01,
        std_max: Number = 1,
        std_perturb_std: Number = 0.1,
        correlation_min: Number = 0.01,
        correlation_max: Number = 1,
        correlation_perturb_std: Number = 0.1,
    ):
        self._name = name
        self.dobs = np.array(dobs)
        self.noise_is_correlated = noise_is_correlated
        self.std_min = std_min
        self.std_max = std_max
        self.std_perturb_std = std_perturb_std
        self.correlation_min = correlation_min
        self.correlation_max = correlation_max
        self.correlation_perturb_std = correlation_perturb_std
        assert std_min >= 0, "standard deviation should always be positive"
        if covariance_mat_inv is not None:
            self._perturbation_func = None
            if np.isscalar(covariance_mat_inv):
                self.covariance_mat_inv = covariance_mat_inv
            else:
                self.covariance_mat_inv = np.array(covariance_mat_inv)
        else:
            self.covariance_mat_inv = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_hierarchical(self):
        """whether the data noise is unknown (i.e. to be inverted for)"""
        return self.covariance_mat_inv is None

    def initialize(self, state: State):
        """initializes the data noise parameters

        Parameters
        ----------
        state : State
            the current state in the Bayesian inference, in which DataNoiseState 
            is to be set
        """
        if self.is_hierarchical:
            noise_std = random.uniform(self.std_min, self.std_max)
            # state.set_param_values((self.name, "noise_std"), noise_std)
            noise_corr = random.uniform(self.correlation_min, self.correlation_max) \
                if self.noise_is_correlated else None
            state.set_param_values(self.name, DataNoiseState(std=noise_std, correlation=noise_corr))

    def inverse_covariance_times_vector(
        self, state: State, vector: np.ndarray
    ) -> np.ndarray:
        """calculates the dot product of the covariance inverse matrix with a given
        vector

        Parameters
        ----------
        state : State
            the current state state
        vector : np.ndarray
            the vector to apply the dot product on

        Returns
        -------
        np.ndarray
            the result from the dot product operation
        """
        if not self.is_hierarchical:
            if np.isscalar(self.covariance_mat_inv):
                return self.covariance_mat_inv * vector
            else:
                return self.covariance_mat_inv @ vector
        else:
            noise = state[self.name]
            std = noise.std
            correlation = noise.correlation
            if correlation is None:
                return 1 / std**2 * vector
            else:
                n = self.dobs.size
                mat = inverse_covariance(std, correlation, n)
                return mat @ vector

    def log_determinant_covariance(self, state: State) -> float:
        r"""the log of the determinant of the covariance matrix
        
        The determinant of the data covariance matrix is calculated assuming
        an exponential decay in the noise correlation between adjacent data 
        points [1]_, i.e.,
        
        .. math::
            \lvert \mathbf{C}_e \rvert = \sigma^{2n} (1 - r^2)^{n-1},
            
        where :math:`\sigma` denotes the standard deviation, :math:`r` the
        correlation, and :math:`n` the size of the data vector.
        

        Parameters
        ----------
        state : State
            the current Bayesian inferernce state

        Returns
        -------
        float
            the log of the determinant
            
        References
        ----------
        .. [1] Bodin et al. 2012, Transdimensional inversion of receiver functions 
            and surface wave dispersion.
        """
        noise = state[self.name]
        std = noise.std
        r = noise.correlation
        if r is None:
            r = 0
        n = self.dobs.size
        log_det = (2 * n) * math.log(std) + (n - 1) * math.log(1 - r**2)
        return log_det

    def __repr__(self) -> str:
        return self.name



# cython: embedsignature=True

cimport cython
from cython cimport cdivision, boundscheck, wraparound
from libcpp cimport bool as bool_cpp
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.algorithm cimport sort
from libc.stdlib cimport malloc, free
from libc.math cimport fabs
import numpy as np
cimport numpy as np
    


@boundscheck(False)
@wraparound(False) 
cpdef bool_cpp is_sorted(double[:] arr):
    cdef long i
    cdef size_t size = arr.shape[0]
    for i in range(size - 1):
        if arr[i] > arr[i + 1]:
            return False
    return True


@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
cpdef compute_voronoi1d_cell_extents(double[:] depth, 
                                     double lb=0, 
                                     double ub=-1,
                                     double fill_value=0):
    cdef size_t size = depth.shape[0]
    cdef int i
    cdef double d1, d2
    cdef double[:] thickness = np.zeros(size, dtype=np.double)
    d1 = lb
    for i in range(size - 1):
        d2 = (depth[i] + depth[i+1]) / 2.
        thickness[i] = d2 - d1 if d1>=0 else fill_value
        d1 = d2
    thickness[i+1] = ub - d2 if ub>=0 else fill_value
    return np.asarray(thickness)


@boundscheck(False)
@wraparound(False) 
cpdef (int, int) _closest_and_final_index(double[:] ndarray, double value):
    cdef int closest_idx = 0 
    cdef int i
    cdef double vmin = fabs(ndarray[0] - value)
    cdef double v
    cdef size_t size = ndarray.shape[0]
    
    for i in range(1, size):
        v = fabs(ndarray[i] - value)
        if v < vmin:
            vmin = v
            closest_idx = i
    if ndarray[closest_idx] > value:
        return closest_idx, closest_idx
    return closest_idx, closest_idx+1
        

@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
cdef double _interpolate_linear_1d(double xp, double[:] x, double[:] y):
    cdef size_t xlen = x.shape[0]
    cdef int i
    cdef double x0, x1, y0, y1
    
    if xp <= x[0]:
        return y[0]
    elif xp >= x[xlen-1]:
        return y[xlen-1]
    i = floor_index(xp, x, xlen)
    x0 = x[i]
    x1 = x[i+1]
    y0 = y[i]
    y1 = y[i+1]
    return y0 + (xp - x0) * (y1 - y0) / (x1 - x0)


@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
cdef double _interpolate_nearest_1d(double xp, double[:] x, double[:] y):
    cdef ssize_t xlen = x.shape[0]
    cdef int i
    cdef double x0, x1, y0, y1
    
    if xp <= x[0]:
        return y[0]
    elif xp >= x[xlen-1]:
        return y[xlen-1]
    i = nearest_index(xp, x, xlen)
    return y[i]


@boundscheck(False)
@wraparound(False) 
cpdef int nearest_index(double xp, double[:] x, ssize_t xlen):
    cdef int i
    cdef double x0, x1
    if xlen==1 or xp<x[0]:
        return 0
    for i in range(xlen - 1):
        x0 = x[i]
        x1 = x[i + 1]
        if x0 <= xp <= x1:
            if fabs(x0-xp) < fabs(x1-xp):
                return i
            return i + 1
    return i + 1


@boundscheck(False)
@wraparound(False) 
cdef int floor_index(double xp, double[:] x, ssize_t xlen):
    cdef int i
    for i in range(xlen):
        if x[i] <= xp <= x[i+1]:
            return i


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_depth_profile(double[:] x, double[:] y, double[:] x0):
    cdef size_t ilayer = 0
    cdef size_t i
    cdef size_t x0_len = x0.shape[0]
    cdef size_t x_len = x.shape[0]
    cdef double interface_lower = 0
    cdef double interface_upper = x[0]
    cdef double x_i
    cdef double[:] ynew = np.zeros(x0_len, dtype=np.double)
    
    for i in range(x0_len):
        x_i = x0[i]
        while True:
            if interface_lower <= x_i < interface_upper or \
                ilayer + 1 >= x_len:
                ynew[i] = y[ilayer]
                break
            else:
                ilayer += 1
                interface_lower = interface_upper
                interface_upper += x[ilayer]
    return np.asarray(ynew)


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_linear_1d(xp, double[:] x, double[:] y):
    """ Linear interpolation in 1-D
    Parameters
    ----------
    xp : float, ndarray of floats
        Interpolation point(s)
        
    x, y : ndarray of floats
        Discrete set of known data points
    
    Returns
    -------
    yp : float
        Interpolated value(s)
    """
    if np.isscalar(xp):
        return _interpolate_linear_1d(xp, x, y)
    cdef int i, size = xp.shape[0]
    cdef double[:] yp = np.zeros(size, dtype=np.double)
    for i in range(size):
        yp[i] = _interpolate_linear_1d(xp[i], x, y)
    return np.asarray(yp)


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_nearest_1d(xp, double[:] x, double[:] y):
    """ Nearest-neighbour interpolation in 1-D
    Parameters
    ----------
    xp : float, ndarray of floats
        Interpolation point(s)
        
    x, y : ndarray of floats
        Discrete set of known data points
    
    Returns
    -------
    yp : float
        Interpolated value(s)
    """
    if np.isscalar(xp):
        return _interpolate_nearest_1d(xp, x, y)
    cdef int i, size = xp.shape[0]
    cdef double[:] yp = np.zeros(size, dtype=np.double)
    for i in range(size):
        yp[i] = _interpolate_nearest_1d(xp[i], x, y)
    return np.asarray(yp)


@boundscheck(False)
@wraparound(False) 
@cdivision(True)
cpdef inverse_covariance(double sigma, double r, size_t n):
    cdef size_t i
    cdef double factor = 1 / (sigma**2 * (1 - r**2))
    cdef double[:, ::1] matrix = np.zeros((n, n), dtype=np.double)
    
    for i in range(1, n-1):
        matrix[i, i] = (1 + r**2) * factor
        matrix[i, i-1] = -r * factor
        matrix[i-1, i] = -r * factor
    matrix[0, 0] = factor
    matrix[n-1, n-1] = factor
    matrix[i+1, i] = -r * factor
    matrix[i, i+1] = -r * factor
    return np.asarray(matrix)
    

@boundscheck(False)
@wraparound(False) 
cpdef insert_scalar(double[:] values, long index, double value):
    cdef size_t new_size = values.shape[0] + 1
    cdef double[:] new_values = np.zeros(new_size, dtype=np.double)
    cdef size_t i = 0
    cdef size_t j = 0
    while i < new_size:
        if i == index:
            new_values[i] = value
        else:
            new_values[i] = values[j]
            j += 1
        i += 1
    return np.asarray(new_values)


@boundscheck(False)
@wraparound(False) 
cpdef delete(double[:] values, long index):
    cdef size_t size = values.shape[0]
    cdef double[:] new_values = np.zeros(size - 1, dtype=np.double)
    cdef size_t i, j = 0
    for i in range(size):
        if i != index:
            new_values[j] = values[i]
            j += 1
    return np.asarray(new_values)

